// Copyright (c) Microsoft Corporation. All rights reserved.

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.TeamFoundation.VersionControl.Common;
using System.Globalization;
using Microsoft.TeamFoundation;
using Microsoft.TeamFoundation.Migration.Toolkit.VC;
using System.Diagnostics;

namespace Microsoft.TeamFoundation.Migration.Toolkit
{
    public class ChangeOptimizer
    {
        List<BatchedItem> m_resolvedChanges = new List<BatchedItem>();
        List<BatchedItem> m_unresolvedChanges = new List<BatchedItem>();
        object m_locker = new object();

        bool acceptingNewChanges = true;

        public void Clear()
        {
            lock (m_locker)
            {
                m_resolvedChanges.Clear();
                m_unresolvedChanges.Clear();
                acceptingNewChanges = true;
            }
        }

        public void Add(BatchedItem change)
        {
            if (change == null)
            {
                throw new ArgumentNullException("change");
            }

            lock (m_locker)
            {
                if (!acceptingNewChanges)
                {
                    throw new InvalidOperationException("Not accepting new changes");
                }

                // Don't add undelete, rename or delete actions if the parent folders of the same actions have already been added.
                if ((change.Action == ChangeAction.Rename) || (change.Action == ChangeAction.Undelete) || (change.Action == ChangeAction.Delete))
                {
                    for (int i = m_unresolvedChanges.Count - 1; i >= 0; i--)
                    {
                        if (TfsUtil.isChildItemOf(change, m_unresolvedChanges[i]))
                        {
                            // Parent item has already been added, skip the add and return. 
                            return;
                        }
                        if (TfsUtil.isChildItemOf(m_unresolvedChanges[i], change))
                        {
                            // Child item has already been added, remove the child item and continue
                            m_unresolvedChanges.RemoveAt(i);
                        }
                    }
                }
                m_unresolvedChanges.Add(change);
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity")]
        private void detectConflicts()
        {
            for (int outerIndex = 0; outerIndex < m_unresolvedChanges.Count; outerIndex++)
            {
                BatchedItem processedChange = m_unresolvedChanges[outerIndex];

                for (int innerIndex = 0; innerIndex < m_unresolvedChanges.Count; innerIndex++)
                {
                    if (innerIndex == outerIndex)
                    {
                        continue;
                    }
                    BatchedItem change = m_unresolvedChanges[innerIndex];
                    // For two Rename actions, processedChange - A rename to B, change - B rename to C.
                    // We need to execute change first. processedChange.ConflictItem = change
                    if (processedChange.Target != null &&
                        change.Source != null &&
                        change.Action == ChangeAction.Rename &&
                        processedChange.Action != ChangeAction.Branch &&
                        processedChange.Action != ChangeAction.Merge &&
                        processedChange.Action != ChangeAction.Delete )
                    {
                        if (VersionControlPath.IsSubItem(processedChange.Target, change.Source))
                        {
                            if (processedChange.ConflictItem != null && processedChange.ConflictItem.ID != change.ID)
                            {
                                throw new UnresolvableConflictException("Multiple items conflict with the same change");
                            }

                            processedChange.ConflictItem = change;
                        }
                        else if (VersionControlPath.IsSubItem(processedChange.Target, change.Target) && 
                            (processedChange.Action == ChangeAction.Edit || processedChange.Action == ChangeAction.Add))
                        {
                            processedChange.ConflictItem = change;
                        }
                    }
                    else if (processedChange.Source != null &&
                         processedChange.Action == ChangeAction.Rename &&
                         change.Action != ChangeAction.Branch &&
                         change.Action != ChangeAction.Merge &&
                         change.Action != ChangeAction.Delete )
                    {
                        if (VersionControlPath.IsSubItem(change.Target, processedChange.Source))
                        {
                            if (change.ConflictItem != null && change.ConflictItem.ID != processedChange.ID)
                            {
                                throw new UnresolvableConflictException("Multiple items conflict with the same change");
                            }

                            change.ConflictItem = processedChange;
                        }
                        else if (VersionControlPath.IsSubItem(change.Target, processedChange.Target) && 
                            (change.Action == ChangeAction.Edit || change.Action == ChangeAction.Add))
                        {
                            change.ConflictItem = processedChange;
                        }
                    }
                    else if (change.Target != null && processedChange.Target != null &&
                        VersionControlPath.EqualsCaseSensitive(change.Target, processedChange.Target))
                    {
                        if (isAdditiveAction(change) && isAdditiveAction(processedChange))
                        {
                            if (TFStringComparer.FilePath.Compare(change.Target, processedChange.Target) == 0)
                            {
                                // if both actions create a new namespace then we're boned.  But if one alters it and the other creates
                                // it then we want to execute the creational action first.
                                throw new UnresolvableConflictException(
                                    string.Format(MigrationToolkitVCResources.Culture, MigrationToolkitVCResources.TwoChangesWithSameTarget, change.Source)
                                    );
                            }
                        }
                        else
                        {
                            if (isAdditiveAction(change))
                            {
                                processedChange.ConflictItem = change;
                            }
                            else if (isAdditiveAction(processedChange))
                            {
                                change.ConflictItem = processedChange;
                            }
                        }

                    }

                    else if (change.Source != null && processedChange.Source != null &&
                        VersionControlPath.EqualsCaseSensitive(change.Source, processedChange.Source) && 
                        change.Action == ChangeAction.Rename && 
                        processedChange.Action == ChangeAction.Rename)
                    {

                        throw new UnresolvableConflictException(
                            string.Format(MigrationToolkitVCResources.Culture, MigrationToolkitVCResources.TwoChangesWithSameSource, change.Source)
                        );
                    }
                }
            }
        }

        static bool isAdditiveAction(BatchedItem item)
        {
            switch (item.Action)
            {
                case ChangeAction.Add:
                case ChangeAction.Branch:
                case ChangeAction.Rename:
                case ChangeAction.Undelete:
                    return true;
                default:
                    return false;
            }
        }

        public ReadOnlyCollection<BatchedItem> Resolve()
        {
            Debug.Assert(acceptingNewChanges, "this should be true when resolve is called");

            lock (m_locker)
            {
                acceptingNewChanges = false;

                detectConflicts();

                // walk backwards so that RemoveAt can work 
                // without the removals affecting indexing
                for (int i = m_unresolvedChanges.Count - 1; i >= 0; i--)
                {
                    BatchedItem item = m_unresolvedChanges[i];
                    if (item.ConflictItem == null)
                    {
                        item.Resolved = true;
                        m_resolvedChanges.Add(item);
                        m_unresolvedChanges.RemoveAt(i);
                    }
                }

                if (m_unresolvedChanges.Count != 0)
                {
                    int currentCount = m_unresolvedChanges.Count;
                    int prevCount = -1;

                    // detect and remove priority issues
                    while (currentCount > 0 && prevCount != currentCount)
                    {
                        // resolve cases that can be solved just with priority adjustment
                        for (int i = m_unresolvedChanges.Count - 1; i >= 0; i--)
                        {
                            BatchedItem item = m_unresolvedChanges[i];

                            // if the conflict chain ends with the next item
                            // or the next item in the chain is already resolved
                            // then resolve by bumping our priority
                            if (item.ConflictItem.ConflictItem == null ||
                                item.ConflictItem.Resolved)
                            {
                                if ((item.Action == ChangeAction.Add) && (item.ConflictItem.Action == ChangeAction.Rename) && 
                                    VersionControlPath.IsSubItem(item.ConflictItem.Target, item.Target))
                                {
                                    // For the case Add Folder1 and rename Folder1 to Folder1\Folder1, change the action sequence to 
                                    // rename Folder1 to intermediate, Add Folder1, rename intermediate to Folder1\Folder1
                                    string intermediate = getIntermediateName(item.ConflictItem.Source);
                                    BatchedItem intermediateToOriginal =
                                        new BatchedItem(
                                            intermediate,
                                            item.ConflictItem.Target,
                                            ChangeAction.Rename,
                                            item.ConflictItem.Priority + 2);
                                    m_resolvedChanges.Add(intermediateToOriginal);

                                    item.ConflictItem.Target = intermediate;
                                }
                                item.Resolved = true;
                                item.Priority = item.ConflictItem.Priority + 1;
                                m_resolvedChanges.Add(item);
                                m_unresolvedChanges.RemoveAt(i);
                            }
                        }

                        // identify cycles
                        for (int i = 0; i < m_unresolvedChanges.Count; i++)
                        {
                            BatchedItem item = m_unresolvedChanges[i];
                            BatchedItem next = item.ConflictItem;
                            while (next != null)
                            {
                                if (next.ConflictItem != null &&
                                    next.ConflictItem.ID == item.ID)
                                {
                                    breakCycle(item);
                                    break;
                                }

                                next = next.ConflictItem;
                            }
                        }

                        prevCount = currentCount;
                        currentCount = m_unresolvedChanges.Count;
                    }
                }

                if (m_unresolvedChanges.Count != 0)
                {
                    throw new UnresolvableConflictException("Unable to resolve pending changes");
                }

                int highestPriority = 0;
                foreach (BatchedItem item in m_resolvedChanges)
                {
                    if (item.Priority > highestPriority)
                    {
                        highestPriority = item.Priority;
                    }
                }
                foreach (BatchedItem item in m_resolvedChanges)
                {
                    // Delay all renames to be processed at last.
                    if (item.Action == ChangeAction.Rename)
                    {
                        item.Priority = item.Priority + highestPriority + 1;
                    }   
                    // If an item depends on a rename action. Delay this action too.
                    else if ((item.ConflictItem != null) && (item.ConflictItem.Action == ChangeAction.Rename))
                    {
                        item.Priority = item.Priority + highestPriority + 1;
                    }
                }

                m_resolvedChanges.Sort(
                    delegate(BatchedItem lhs, BatchedItem rhs)
                    {
                        return lhs.Priority.CompareTo(rhs.Priority);
                    }
                );

                return new ReadOnlyCollection<BatchedItem>(m_resolvedChanges);
            }
        }

        private int breakCycle(BatchedItem item)
        {
            /*
             *  rename A B
                rename B A
                
             *  becomes
             * 
                rename A <intermediate>
                rename B A
                rename <intermediate> B

                the conflict is from A to B
             */
            if (item.Action != ChangeAction.Rename)
            {
                throw new UnresolvableConflictException("Don't know how to break non-rename cycles");
            }

            if (item.Resolved)
            {
                return item.Priority;
            }

            string intermediate = getIntermediateName(item.Target);
            BatchedItem intermediate1 =
                new BatchedItem(
                    item.Source,
                    intermediate,
                    ChangeAction.Rename,
                    item.ConflictItem.Priority);

            item.ConflictItem.Priority++;

            BatchedItem intermediate2 =
                new BatchedItem(
                    intermediate,
                    item.Target,
                    ChangeAction.Rename,
                    item.ConflictItem.Priority + 1);

            m_resolvedChanges.Add(intermediate1);
            m_resolvedChanges.Add(intermediate2);

            // remove the old change
            item.Resolved = true;
            m_unresolvedChanges.Remove(item);

            if (item.ConflictItem.ConflictItem != null &&
                item.ConflictItem.ConflictItem.ID == item.ID)
            {
                item.ConflictItem.Resolved = true;
                m_unresolvedChanges.Remove(item.ConflictItem);
                m_resolvedChanges.Add(item.ConflictItem);
            }
            else
            {
                intermediate2.Priority = breakCycle(item.ConflictItem);
            }

            return item.ConflictItem.Priority;
        }

        private static string getIntermediateName(string targetName)
        {
            return targetName + Guid.NewGuid().ToString();
        }
    }
}
